@@ -42,7 +42,7 @@ module Grape
   #     end
   #   end
   class Entity
-    attr_reader :object, :options
+    attr_reader :object, :delegator, :options
 
     # The Entity DSL allows you to mix entity functionality into
     # your existing classes.
@@ -98,6 +98,29 @@ module Grape
       end
     end
 
+    class << self
+      def root_exposure
+        @root_exposure ||= Exposure.new(nil, nesting: true)
+      end
+
+      attr_writer :root_exposure
+
+      # Returns all formatters that are registered for this and it's ancestors
+      # @return [Hash] of formatters
+      def formatters
+        @formatters ||= {}
+      end
+
+      attr_writer :formatters
+    end
+
+    @formatters = {}
+
+    def self.inherited(subclass)
+      subclass.root_exposure = root_exposure.dup
+      subclass.formatters = formatters.dup
+    end
+
     # This method is the primary means by which you will declare what attributes
     # should be exposed by the entity.
     #
@@ -133,33 +156,64 @@ module Grape
 
       fail ArgumentError, 'You may not use block-setting when also using format_with' if block_given? && options[:format_with].respond_to?(:call)
 
-      options[:proc] = block if block_given? && block.parameters.any?
+      if block_given?
+        if block.parameters.any?
+          options[:proc] = block
+        else
+          options[:nesting] = true
+        end
+      end
 
-      @nested_attributes ||= []
+      @documentation = nil
+      @nesting_stack ||= []
 
+      # rubocop:disable Style/Next
       args.each do |attribute|
-        unless @nested_attributes.empty?
-          orig_attribute = attribute.to_sym
-          attribute = "#{@nested_attributes.last}__#{attribute}"
-          nested_attribute_names_hash[attribute.to_sym] = orig_attribute
-          options[:nested] = true
-          nested_exposures_hash[@nested_attributes.last.to_sym] ||= {}
-          nested_exposures_hash[@nested_attributes.last.to_sym][attribute.to_sym] = options
+        exposure = Exposure.new(attribute, options)
+
+        if @nesting_stack.empty?
+          root_exposures << exposure
+        else
+          @nesting_stack.last.nested_exposures << exposure
         end
 
-        exposures[attribute.to_sym] = options
-
         # Nested exposures are given in a block with no parameters.
-        if block_given? && block.parameters.empty?
-          @nested_attributes << attribute
+        if exposure.nesting?
+          @nesting_stack << exposure
           block.call
-          @nested_attributes.pop
+          @nesting_stack.pop
         end
       end
     end
 
-    def self.unexpose(attribute)
-      exposures.delete(attribute)
+    # Returns exposures that have been declared for this Entity on the top level.
+    # @return [Array] of exposures
+    def self.root_exposures
+      root_exposure.nested_exposures
+    end
+
+    def self.find_exposure(attribute)
+      root_exposures.find_by(attribute)
+    end
+
+    def self.unexpose(*attributes)
+      cannot_unexpose! unless can_unexpose?
+      @documentation = nil
+      root_exposures.delete_by(*attributes)
+    end
+
+    def self.unexpose_all
+      cannot_unexpose! unless can_unexpose?
+      @documentation = nil
+      root_exposures.clear
+    end
+
+    def self.can_unexpose?
+      (@nesting_stack ||= []).empty?
+    end
+
+    def self.cannot_unexpose!
+      fail "You cannot call 'unexpose` inside of nesting exposure!"
     end
 
     # Set options that will be applied to any exposures declared inside the block.
@@ -177,74 +231,15 @@ module Grape
       @block_options.pop
     end
 
-    # Returns a hash of exposures that have been declared for this Entity or ancestors. The keys
-    # are symbolized references to methods on the containing object, the values are
-    # the options that were passed into expose.
-    def self.exposures
-      return @exposures unless @exposures.nil?
-
-      @exposures = {}
-
-      if superclass.respond_to? :exposures
-        @exposures = superclass.exposures.merge(@exposures)
-      end
-
-      @exposures
-    end
-
-    class << self
-      attr_accessor :_nested_attribute_names_hash
-      attr_accessor :_nested_exposures_hash
-
-      def nested_attribute_names_hash
-        self._nested_attribute_names_hash ||= {}
-      end
-
-      def nested_exposures_hash
-        self._nested_exposures_hash ||= {}
-      end
-
-      def nested_attribute_names
-        return @nested_attribute_names unless @nested_attribute_names.nil?
-
-        @nested_attribute_names = {}.merge(nested_attribute_names_hash)
-
-        if superclass.respond_to? :nested_attribute_names
-          @nested_attribute_names = superclass.nested_attribute_names.deep_merge(@nested_attribute_names)
-        end
-
-        @nested_attribute_names
-      end
-
-      def nested_exposures
-        return @nested_exposures unless @nested_exposures.nil?
-
-        @nested_exposures = {}.merge(nested_exposures_hash)
-
-        if superclass.respond_to? :nested_exposures
-          @nested_exposures = superclass.nested_exposures.deep_merge(@nested_exposures)
-        end
-
-        @nested_exposures
-      end
-    end
-
     # Returns a hash, the keys are symbolized references to fields in the entity,
     # the values are document keys in the entity's documentation key. When calling
     # #docmentation, any exposure without a documentation key will be ignored.
     def self.documentation
-      @documentation ||= exposures.inject({}) do |memo, (attribute, exposure_options)|
-        unless exposure_options[:documentation].nil? || exposure_options[:documentation].empty?
-          memo[key_for(attribute)] = exposure_options[:documentation]
+      @documentation ||= root_exposures.each_with_object({}) do |exposure, memo|
+        if exposure.documentation && !exposure.documentation.empty?
+          memo[exposure.key] = exposure.documentation
         end
-        memo
       end
-
-      if superclass.respond_to? :documentation
-        @documentation = superclass.documentation.merge(@documentation)
-      end
-
-      @documentation
     end
 
     # This allows you to declare a Proc in which exposures can be formatted with.
@@ -278,17 +273,6 @@ module Grape
       formatters[name.to_sym] = block
     end
 
-    # Returns a hash of all formatters that are registered for this and it's ancestors.
-    def self.formatters
-      @formatters ||= {}
-
-      if superclass.respond_to? :formatters
-        @formatters = superclass.formatters.merge(@formatters)
-      end
-
-      @formatters
-    end
-
     # This allows you to set a root element name for your representation.
     #
     # @param plural   [String] the root key to use when representing
@@ -352,7 +336,7 @@ module Grape
     #
     #       class Users < Grape::Entity
     #         present_collection true
-    #         expose :items, as: 'users', using: API::Entities::Users
+    #         expose :items, as: 'users', using: API::Entities::User
     #         expose :version, documentation: { type: 'string',
     #                                           desc: 'actual api version',
     #                                           required: true }
@@ -404,13 +388,15 @@ module Grape
     #   even if one is defined for the entity.
     # @option options :serializable [true or false] when true a serializable Hash will be returned
     #
+    # @option options :only [Array] all the fields that should be returned
+    # @option options :except [Array] all the fields that should not be returned
     def self.represent(objects, options = {})
       if objects.respond_to?(:to_ary) && ! @present_collection
-        root_element =  @collection_root
-        inner = objects.to_ary.map { |object| new(object, { collection: true }.merge(options)).presented }
+        root_element =  root_element(:collection_root)
+        inner = objects.to_ary.map { |object| new(object, options.reverse_merge(collection: true)).presented }
       else
         objects = { @collection_name => objects } if @present_collection
-        root_element =  @root
+        root_element = root_element(:root)
         inner = new(objects, options).presented
       end
 
@@ -419,6 +405,16 @@ module Grape
       root_element ? { root_element => inner } : inner
     end
 
+    # This method returns the entity's root or collection root node, or its parent's
+    # @param root_type: either :collection_root or just :root
+    def self.root_element(root_type)
+      if instance_variable_get("@#{root_type}")
+        instance_variable_get("@#{root_type}")
+      elsif superclass.respond_to? :root_element
+        superclass.root_element(root_type)
+      end
+    end
+
     def presented
       if options[:serializable]
         serializable_hash
@@ -428,17 +424,21 @@ module Grape
     end
 
     def initialize(object, options = {})
-      @object, @options = object, options
+      @object = object
+      @delegator = Delegator.new object
+      @options = if options.is_a? Options
+                   options
+                 else
+                   Options.new options
+                 end
     end
 
-    def exposures
-      self.class.exposures
+    def root_exposures
+      self.class.root_exposures
     end
 
-    def valid_exposures
-      exposures.reject { |_, options| options[:nested] }.select do |attribute, exposure_options|
-        valid_exposure?(attribute, exposure_options)
-      end
+    def root_exposure
+      self.class.root_exposure
     end
 
     def documentation
@@ -458,24 +458,29 @@ module Grape
     #   etc.
     def serializable_hash(runtime_options = {})
       return nil if object.nil?
+
       opts = options.merge(runtime_options || {})
-      valid_exposures.inject({}) do |output, (attribute, exposure_options)|
-        if conditions_met?(exposure_options, opts)
-          partial_output = value_for(attribute, opts)
-          output[self.class.key_for(attribute)] =
-            if partial_output.respond_to? :serializable_hash
-              partial_output.serializable_hash(runtime_options)
-            elsif partial_output.is_a?(Array) && !partial_output.map { |o| o.respond_to? :serializable_hash }.include?(false)
-              partial_output.map(&:serializable_hash)
-            elsif partial_output.is_a?(Hash)
-              partial_output.each do |key, value|
-                partial_output[key] = value.serializable_hash if value.respond_to? :serializable_hash
-              end
-            else
-              partial_output
-            end
-        end
-        output
+
+      root_exposure.serializable_value(self, opts)
+    end
+
+    def exec_with_object(options, &block)
+      instance_exec(object, options, &block)
+    end
+
+    def exec_with_attribute(attribute, &block)
+      instance_exec(delegate_attribute(attribute), &block)
+    end
+
+    def value_for(key, options = Options.new)
+      root_exposure.valid_value_for(key, self, options)
+    end
+
+    def delegate_attribute(attribute)
+      if respond_to?(attribute, true)
+        send(attribute)
+      else
+        delegator.delegate(attribute)
       end
     end
 
@@ -491,130 +496,9 @@ module Grape
       serializable_hash(options).to_xml(options)
     end
 
-    protected
-
-    def self.name_for(attribute)
-      attribute = attribute.to_sym
-      nested_attribute_names[attribute] || attribute
-    end
-
-    def self.key_for(attribute)
-      exposures[attribute.to_sym][:as] || name_for(attribute)
-    end
-
-    def self.nested_exposures_for(attribute)
-      nested_exposures[attribute] || {}
-    end
-
-    def value_for(attribute, options = {})
-      exposure_options = exposures[attribute.to_sym]
-
-      nested_exposures = self.class.nested_exposures_for(attribute)
-
-      if exposure_options[:using]
-        exposure_options[:using] = exposure_options[:using].constantize if exposure_options[:using].respond_to? :constantize
-
-        using_options = options.dup
-        using_options.delete(:collection)
-        using_options[:root] = nil
-
-        if exposure_options[:proc]
-          exposure_options[:using].represent(instance_exec(object, options, &exposure_options[:proc]), using_options)
-        else
-          exposure_options[:using].represent(delegate_attribute(attribute), using_options)
-        end
-
-      elsif exposure_options[:proc]
-        instance_exec(object, options, &exposure_options[:proc])
-
-      elsif exposure_options[:format_with]
-        format_with = exposure_options[:format_with]
-
-        if format_with.is_a?(Symbol) && formatters[format_with]
-          instance_exec(delegate_attribute(attribute), &formatters[format_with])
-        elsif format_with.is_a?(Symbol)
-          send(format_with, delegate_attribute(attribute))
-        elsif format_with.respond_to? :call
-          instance_exec(delegate_attribute(attribute), &format_with)
-        end
-
-      elsif nested_exposures.any?
-        nested_attributes =
-          nested_exposures.map do |nested_attribute, nested_exposure_options|
-            if conditions_met?(nested_exposure_options, options)
-              [self.class.key_for(nested_attribute), value_for(nested_attribute, options)]
-            end
-          end
-
-        Hash[nested_attributes.compact]
-      else
-        delegate_attribute(attribute)
-      end
-    end
-
-    def delegate_attribute(attribute)
-      name = self.class.name_for(attribute)
-      if respond_to?(name, true)
-        send(name)
-      elsif object.is_a?(Hash)
-        object[name]
-      elsif object.respond_to?(name, true)
-        object.send(name)
-      elsif object.respond_to?(:fetch, true)
-        object.fetch(name)
-      else
-        begin
-          object.send(name)
-        rescue NoMethodError
-          raise NoMethodError, "#{self.class.name} missing attribute `#{name}' on #{object}"
-        end
-      end
-    end
-
-    def valid_exposure?(attribute, exposure_options)
-      nested_exposures = self.class.nested_exposures_for(attribute)
-      (nested_exposures.any? && nested_exposures.all? { |a, o| valid_exposure?(a, o) }) || \
-        exposure_options.key?(:proc) || \
-        !exposure_options[:safe] || \
-        object.respond_to?(self.class.name_for(attribute)) || \
-        object.is_a?(Hash) && object.key?(self.class.name_for(attribute))
-    end
-
-    def conditions_met?(exposure_options, options)
-      if_conditions = []
-      unless exposure_options[:if_extras].nil?
-        if_conditions.concat(exposure_options[:if_extras])
-      end
-      if_conditions << exposure_options[:if] unless exposure_options[:if].nil?
-
-      if_conditions.each do |if_condition|
-        case if_condition
-        when Hash then if_condition.each_pair { |k, v| return false if options[k.to_sym] != v }
-        when Proc then return false unless instance_exec(object, options, &if_condition)
-        when Symbol then return false unless options[if_condition]
-        end
-      end
-
-      unless_conditions = []
-      unless exposure_options[:unless_extras].nil?
-        unless_conditions.concat(exposure_options[:unless_extras])
-      end
-      unless_conditions << exposure_options[:unless] unless exposure_options[:unless].nil?
-
-      unless_conditions.each do |unless_condition|
-        case unless_condition
-        when Hash then unless_condition.each_pair { |k, v| return false if options[k.to_sym] == v }
-        when Proc then return false if instance_exec(object, options, &unless_condition)
-        when Symbol then return false if options[unless_condition]
-        end
-      end
-
-      true
-    end
-
     # All supported options.
     OPTIONS = [
-      :as, :if, :unless, :using, :with, :proc, :documentation, :format_with, :safe, :if_extras, :unless_extras
+      :rewrite, :as, :if, :unless, :using, :with, :proc, :documentation, :format_with, :safe, :attr_path, :if_extras, :unless_extras
     ].to_set.freeze
 
     # Merges the given options with current block options.
